#include "VNode.h"
#include "../bitcode/BitcodeMod.hpp"

#include <cassert>
#include <csignal>
#include <cstdlib>
#include <cstring>
#include <exception>
#include <string>
#include <unistd.h>
#include <iostream>

#include <llvm/Module.h>
#include <llvm/Function.h>
#include <llvm/Type.h>
#include <llvm/DerivedTypes.h>
#include <llvm/ModuleProvider.h>
#include <llvm/Bitcode/ReaderWriter.h>
#include <llvm/System/DynamicLibrary.h>
#include <llvm/Support/MemoryBuffer.h>
#include <llvm/ExecutionEngine/ExecutionEngine.h>
#include <llvm/ExecutionEngine/GenericValue.h>

static inline void load_lib(const char* so_name) {
	std::string err;

	std::cout << "loading \"" << so_name << "\"... ";
	if (llvm::sys::DynamicLibrary::LoadLibraryPermanently(so_name, &err)) {
		std::cout << err << ".\n";
		exit(1);
	}
	std::cout << "done.\n";
}

static inline llvm::GenericValue runFunction_ZeroArg(const char* fn, llvm::ExecutionEngine* ee, const llvm::Module* mod) {
	std::cout << "searching \"" << fn << "\"... ";
	llvm::Function* F = mod -> getFunction(fn);
		
	if (F == NULL) {
		std::cout << "not found.\n";
		exit(1);
	}
	std::cout << "found.\n";

	std::cout << "running \"" << F -> getName() << "\" ... ";
	std::vector<llvm::GenericValue> args;
	llvm::GenericValue gv = ee -> runFunction(F, args);
	std::cout << "done.\n";

	return gv;
}

static inline llvm::GenericValue runMemberFunction(llvm::GenericValue& thisPtr, const char* fn, llvm::ExecutionEngine* ee, const llvm::Module* mod) {
	std::cout << "searching \"" << fn << "\"... ";
	llvm::Function* F = mod -> getFunction(fn);

	if (F == NULL) {
		std::cout << "not found.\n";
		exit(1);
	}
	std::cout << "found.\n";

// 	std::cout << "running \"" << F -> getName() << "\" ... ";
	std::vector<llvm::GenericValue> args;
	args.push_back(thisPtr);
	llvm::GenericValue gv = ee -> runFunction(F, args);
// 	std::cout << "done.\n";

	return gv;
}


static inline llvm::GenericValue load_bc(const char* bc_name, const char* so_name) {
	std::string err;
	llvm::GenericValue gv;

	load_lib(so_name);

	std::cout << "reading \"" << bc_name << "\"... ";
	llvm::MemoryBuffer* mb = llvm::MemoryBuffer::getFile(bc_name, &err);
	if (mb == NULL) {
		std::cout << err << ".\n";
		exit(1);
	}

	llvm::ModuleProvider* mp = llvm::getBitcodeModuleProvider(mb, &err);
	if (mp == NULL) {
		std::cout << err << ".\n";
		exit(1);
	}
	std::cout << "done.\n";

	std::cout << "extracting module... ";
	llvm::Module *mod = mp -> getModule();
	if (!mod) {
		std::cout << err << "\n";
		exit(1);
	}
	std::cout << "done.\n";

	std::cout << "creating llvm::ExecutionEngine... ";
	llvm::ExecutionEngine* ee = llvm::ExecutionEngine::create(mp, false, &err);
	if (ee == NULL) {
		std::cout << err << ".\n";
		exit(1);
	}
	std::cout << "done.\n";

	std::cout << "running static constructors... ";
	ee->runStaticConstructorsDestructors(false);
	std::cout << "done.\n";

	gv = runFunction_ZeroArg("_ZN10BitcodeMod6createEv", ee, mod);
	runMemberFunction(gv, "_ZN10BitcodeMod9initVNodeEv", ee, mod);

	return gv;
}

int main(int argc, char* argv[])
{
	BitcodeMod* vnode = NULL;
	assert(argc == 3 && "LLVM_JIT_Test <path to bc> <path to libcore.so>");
	vnode = static_cast<BitcodeMod*>(load_bc(argv[1], argv[2]).PointerVal);

	if (vnode != 0) {
		std::cout << "1\n";
// 		vnode -> initVNode();
		std::cout << "2\n";
	}

	return 0;
}
